﻿using SharpGL;
using SharpGL.SceneGraph;
using System;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Windows;

namespace libTextSuiteExample
{
    public partial class MainWindow : Window
    {
        private static readonly string TEST_TEXT = @"Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.";

        private lts.Context _context;
        private lts.FontCreator _fontCreator;
        private lts.Renderer _renderer;
        private lts.Font _font;
        private lts.Image _pattern;
        private lts.PostProcessorList _ppList;

        private static readonly string LibTextSuiteDllName = "libtextsuite.dll";

        [DllImport("kernel32.dll")]
        private static extern IntPtr LoadLibrary(string dllToLoad);

        public MainWindow()
        {
            InitializeComponent();

            // load libtextsuite depending on architecture
            string exeDir = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location);
            string archDir = Environment.Is64BitProcess ? "x64" : "x86";
            string dllPath = Path.Combine(exeDir, "resources", archDir, LibTextSuiteDllName);
            var handle = LoadLibrary(dllPath);
            if (handle == IntPtr.Zero)
            { 
                MessageBox.Show("Unable to load libtextsuite.dll", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                Application.Current.Shutdown();
            }
        }
        
        private void oglControl_OpenGLInitialized(object sender, OpenGLEventArgs args)
        {
            var gl = args.OpenGL;
            gl.Disable(OpenGL.GL_DEPTH_TEST);
            gl.Disable(OpenGL.GLU_CULLING);
            gl.Enable(OpenGL.GL_BLEND);
            gl.ClearColor(1.0f, 1.0f, 1.0f, 1.0f);
            gl.MakeCurrent();

            string exeDir = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location);
            string fontPath = Path.Combine(exeDir, "resources", "Prototype.ttf");

            _context = new lts.Context();
            _renderer = new lts.RendererOpenGl(_context, lts.Format.RGBA8);
            _fontCreator = new lts.FontCreatorGdi(_context);
            _font = _fontCreator.GetFontByFile(fontPath, 40, lts.FontStyles.Empty, lts.AntiAliasing.Normal);

            _ppList = new lts.PostProcessorList(_context);
            _font.PostProcessor = _ppList;

            _pattern = new lts.Image(_context);
            _pattern.CreateEmpty(lts.Format.Alpha8, 4, 4);
            _pattern.Data = new byte[] {
                0xFF, 0xBF, 0x7F, 0xBF,
                0xBF, 0xFF, 0xBF, 0x7F,
                0x7F, 0xBF, 0xFF, 0xBF,
                0xBF, 0x7F, 0xBF, 0xFF
            };
            
            lts.PostProcessor pp = new lts.PostProcessorFillPattern(
                _context,
                _pattern,
                new lts.Position() { x = 0, y = 0 },
                lts.Constants.ImageModeModulateAll,
                lts.Constants.ColorChannelsRGBA);
            pp.AddChars(lts.CharRangeUsage.Include, "Lorem");
            _ppList.Add(pp);

            pp = new lts.PostProcessorFillColor(
                _context,
                new lts.Color() { R = 0.0f, G = 0.0f, B = 0.5f, A = 1.0f },
                lts.Constants.ImageModeReplaceAll,
                lts.Constants.ColorChannelsRGB);
            pp.AddChars(lts.CharRangeUsage.Exclude, "e");
            _ppList.Add(pp);

            pp = new lts.PostProcessorBorder(
                _context,
                3.0f,
                0.5f,
                new lts.Color() { R = 0.0f, G = 0.5f, B = 0.0f, A = 1.0f },
                true);
            pp.AddChars(lts.CharRangeUsage.Include, "e");
            _ppList.Add(pp);
        }

        private void oglControl_OpenGLDraw(object sender, OpenGLEventArgs args)
        {
            var gl = args.OpenGL;
            gl.Clear(OpenGL.GL_COLOR_BUFFER_BIT | OpenGL.GL_DEPTH_BUFFER_BIT);
            gl.BlendFunc(OpenGL.GL_SRC_ALPHA, OpenGL.GL_ONE_MINUS_SRC_ALPHA);
            using (var textBlock = _renderer.BeginBlock(10, 10, (Int32)oglControl.ActualWidth - 20, (Int32)oglControl.ActualHeight - 20, lts.BlockFlags.WordWrap))
            {
                textBlock.HorzAlign = lts.HorzAlign.Justify;
                textBlock.Font = _font;
                textBlock.Color = new lts.Color() { R = 1.0f, G = 1.0f, B = 1.0f, A = 1.0f };
                textBlock.TextOut(TEST_TEXT);
                _renderer.EndBlock(textBlock);
            }
            gl.Flush();
        }

        private void oglControl_Resized(object sender, OpenGLEventArgs args)
        {
            var gl = args.OpenGL;
            gl.MatrixMode(OpenGL.GL_PROJECTION);
            gl.LoadIdentity();
            gl.Ortho(0, gl.RenderContextProvider.Width, gl.RenderContextProvider.Height, 0, 10, -10);

            gl.MatrixMode(OpenGL.GL_MODELVIEW);
            gl.LoadIdentity();
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (_ppList != null)
            {
                foreach (var pp in _ppList)
                    pp.Dispose();
                _ppList.Dispose();
                _ppList.Clear();
                _ppList = null;
            }

            if (_pattern != null)
            {
                _pattern.Dispose();
                _pattern = null;
            }

            if (_font != null)
            {
                _font.Dispose();
                _font = null;
            }

            if (_fontCreator != null)
            {
                _fontCreator.Dispose();
                _fontCreator = null;
            }

            if (_renderer != null)
            {
                _renderer.Dispose();
                _renderer = null;
            }

            if (_context != null)
            {
                _context.Dispose();
                _context = null;
            }
        }
    }
}
