unit ultsPostProcessor;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  ultsTypes, utsTextSuite;

type
  TltsPostProcessorExecuteFunc = procedure(const aCharHandle: TltsCharHandle; const aImageHandle: TltsImageHandle; aArgs: Pointer); stdcall;

  PltsPostProcessorCustomData = ^TltsPostProcessorCustomData;
  TltsPostProcessorCustomData = packed record
    args: Pointer;
    execute: TltsPostProcessorExecuteFunc;
  end;

  TltsPostProcessorCustom = class(TtsPostProcessor)
  private
    fData: TltsPostProcessorCustomData;
  public
    function Execute(const aChar: TtsChar; const aImage: TtsImage): Boolean; override;
    constructor Create(const aContext: TtsContext; const aData: TltsPostProcessorCustomData);
  end;

function ltsPostProcessorAddRange           (aHandle: TltsPostProcessorHandle; aUsage: TtsCharRangeUsage; aStart, aStop: WideChar): TltsErrorCode;            stdcall;
function ltsPostProcessorAddChars           (aHandle: TltsPostProcessorHandle; aUsage: TtsCharRangeUsage; aChars: PWideChar):       TltsErrorCode;            stdcall;
function ltsPostProcessorClearRanges        (aHandle: TltsPostProcessorHandle):                                                     TltsErrorCode;            stdcall;
function ltsPostProcessorExecute            (aHandle: TltsPostProcessorHandle; aChar: TltsCharHandle; aImage: TltsImageHandle):     TltsErrorCode;            stdcall;

function ltsPostProcessorFillColorCreate    (aContext: TltsContextHandle; aColor: TtsColor4f;
                                             aModes: TtsImageModes; aChannels: TtsColorChannels):                                   TltsPostProcessorHandle;  stdcall;
function ltsPostProcessorFillPatternCreate  (aContext: TltsContextHandle; aPattern: TltsImageHandle; aOwnsPatter: Boolean;
                                             aPosition: TtsPosition; aModes: TtsImageModes; aChannels: TtsColorChannels):           TltsPostProcessorHandle;  stdcall;
function ltsPostProcessorBorderCreate       (aContext: TltsContextHandle; aWidth, aStrength: Single;
                                             aColor: TtsColor4f; aKeepSize: Boolean):                                               TltsPostProcessorHandle;  stdcall;
function ltsPostProcessorShadowCreate       (aContext: TltsContextHandle; aRadius, aStrength: Single;
                                             aOffset: TtsPosition; aColor: TtsColor4f):                                             TltsPostProcessorHandle;  stdcall;
function ltsPostProcessorCustomCreate       (aContext: TltsContextHandle; constref aData: TltsPostProcessorCustomData):             TltsPostProcessorHandle;  stdcall;
function ltsPostProcessorDestroy            (aHandle: TltsPostProcessorHandle):                                                     TltsErrorCode;            stdcall;

implementation

uses
  ultsUtils {$IFDEF DEBUG}, uutlLogger{$ENDIF};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//ltsPostProcessor//////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorAddRange(aHandle: TltsPostProcessorHandle; aUsage: TtsCharRangeUsage; aStart, aStop: WideChar): TltsErrorCode; stdcall;
var
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorAddRange(Handle=%p; Usage=%d; Start=%d; Stop=%d)', [Pointer(aHandle), Integer(aUsage), Integer(aStart), Integer(aStop)]);{$ENDIF}
    result := ltsErrNone;
    if ValidateCharRangeUsage(aUsage) and CheckPostProcessorHandle(aHandle, TtsPostProcessor, pp)
      then pp.AddRange(aUsage, aStart, aStop)
      else result := LastErrorCode;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := LastErrorCode;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorExecute=%d', [Integer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorAddChars(aHandle: TltsPostProcessorHandle; aUsage: TtsCharRangeUsage; aChars: PWideChar): TltsErrorCode; stdcall;
var
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorAddChars(Handle=%p; Usage=%d; Chars=%p)', [Pointer(aHandle), Integer(aUsage), Pointer(aChars)]);{$ENDIF}
    result := ltsErrNone;
    if ValidateCharRangeUsage(aUsage) and CheckPostProcessorHandle(aHandle, TtsPostProcessor, pp)
      then pp.AddChars(aUsage, aChars)
      else result := LastErrorCode;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := LastErrorCode;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorExecute=%d', [Integer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorClearRanges(aHandle: TltsPostProcessorHandle): TltsErrorCode; stdcall;
var
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorClearRanges(Handle=%p)', [Pointer(aHandle)]);{$ENDIF}
    result := ltsErrNone;
    if CheckPostProcessorHandle(aHandle, TtsPostProcessor, pp)
      then pp.ClearRanges
      else result := LastErrorCode;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := LastErrorCode;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorExecute=%d', [Integer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorExecute(aHandle: TltsPostProcessorHandle; aChar: TltsCharHandle; aImage: TltsImageHandle): TltsErrorCode; stdcall;
var
  pp: TtsPostProcessor;
  c: TtsChar;
  i: TtsImage;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorExecute(Handle=%p; Char=%p; Image=%p)', [Pointer(aHandle), Pointer(aChar), Pointer(aImage)]);{$ENDIF}
    result := ltsErrNone;
    if CheckPostProcessorHandle(aHandle, TtsPostProcessor, pp) and
       CheckCharHandle(aChar, c) and
       CheckImageHandle(aImage, i)
          then pp.Execute(c, i)
          else result := LastErrorCode;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := LastErrorCode;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorExecute=%d', [Integer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorFillColorCreate(aContext: TltsContextHandle; aColor: TtsColor4f; aModes: TtsImageModes; aChannels: TtsColorChannels): TltsPostProcessorHandle; stdcall;
var
  c: TtsContext;
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorFillColorCreate(Context=%p; Color=(%f;%f;%f;%f); Modes=(%d;%d;%d;%d); Channels=%d)', [Pointer(aContext), aColor.r, aColor.g, aColor.b, aColor.a, Integer(aModes[TtsColorChannel.tsChannelRed]), Integer(aModes[TtsColorChannel.tsChannelGreen]), Integer(aModes[TtsColorChannel.tsChannelBlue]), Integer(aModes[TtsColorChannel.tsChannelAlpha]), Integer(aChannels)]);{$ENDIF}
    result := nil;
    if CheckContextHandle(aContext, c) then begin
      pp := TtsPostProcessorFillColor.Create(c, aColor, aModes, aChannels);
      AddReference(ltsObjTypePostProcessor, pp);
      result := pp;
    end;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := nil;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorFillColorCreate=%p', [Pointer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorFillPatternCreate(aContext: TltsContextHandle; aPattern: TltsImageHandle; aOwnsPatter: Boolean; aPosition: TtsPosition; aModes: TtsImageModes; aChannels: TtsColorChannels): TltsPostProcessorHandle; stdcall;
var
  c: TtsContext;
  img: TtsImage;
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorFillPatternCreate(Context=%p; Pattern=%p; OwnsPattern=%d; Position=(%d;%d); Modes=(%d;%d;%d;%d); Channels=%d)', [Pointer(aContext), Pointer(aPattern), Integer(aOwnsPatter), aPosition.x, aPosition.y, Integer(aModes[TtsColorChannel.tsChannelRed]), Integer(aModes[TtsColorChannel.tsChannelGreen]), Integer(aModes[TtsColorChannel.tsChannelBlue]), Integer(aModes[TtsColorChannel.tsChannelAlpha]), Integer(aChannels)]);{$ENDIF}
    result := nil;
    img    := nil;
    if CheckContextHandle(aContext, c) and CheckImageHandle(aPattern, img) then begin
      pp := TtsPostProcessorFillPattern.Create(c, img, aOwnsPatter, aPosition, aModes, aChannels);
      AddReference(ltsObjTypePostProcessor, pp);
      result := pp;
    end;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := nil;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorFillPatternCreate=%p', [Pointer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorBorderCreate(aContext: TltsContextHandle; aWidth, aStrength: Single; aColor: TtsColor4f; aKeepSize: Boolean): TltsPostProcessorHandle; stdcall;
var
  c: TtsContext;
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorBorderCreate(Context=%p; Width=%f; Strength=%f; Color=(%f;%f;%f;%f); KeepSize=%d)', [Pointer(aContext), aWidth, aStrength, aColor.r, aColor.g, aColor.b, aColor.a, Integer(aKeepSize)]);{$ENDIF}
    result := nil;
    if CheckContextHandle(aContext, c) then begin
      pp := TtsPostProcessorBorder.Create(c, aWidth, aStrength, aColor, aKeepSize);
      AddReference(ltsObjTypePostProcessor, pp);
      result := pp;
    end;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := nil;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorBorderCreate=%p', [Pointer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorShadowCreate(aContext: TltsContextHandle; aRadius, aStrength: Single; aOffset: TtsPosition; aColor: TtsColor4f): TltsPostProcessorHandle; stdcall;
var
  c: TtsContext;
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorShadowCreate(Context=%p; Radius=%f; Strength=%f; Offset=(%d;%d); Color=(%f;%f;%f;%f))', [Pointer(aContext), aRadius, aStrength, aOffset.x, aOffset.y, aColor.r, aColor.g, aColor.b, aColor.a]);{$ENDIF}
    result := nil;
    if CheckContextHandle(aContext, c) then begin
      pp := TtsPostProcessorShadow.Create(c, aRadius, aStrength, aOffset, aColor);
      AddReference(ltsObjTypePostProcessor, pp);
      result := pp;
    end;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := nil;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorShadowCreate=%p', [Pointer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorCustomCreate(aContext: TltsContextHandle; constref aData: TltsPostProcessorCustomData): TltsPostProcessorHandle; stdcall;
var
  c: TtsContext;
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorCustomCreate(Context=%p; Data=%p)', [Pointer(aContext), Pointer(@aData)]);{$ENDIF}
    result := nil;
    if CheckContextHandle(aContext, c) then begin
      pp := TltsPostProcessorCustom.Create(c, aData);
      AddReference(ltsObjTypePostProcessor, pp);
      result := pp;
    end;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := nil;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorCustomCreate=%p', [Pointer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ltsPostProcessorDestroy(aHandle: TltsPostProcessorHandle): TltsErrorCode; stdcall;
var
  pp: TtsPostProcessor;
begin
  try
    {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorDestroy(Context=%p)', [Pointer(aHandle)]);{$ENDIF}
    result := ltsErrNone;
    if CheckPostProcessorHandle(aHandle, TtsPostProcessor, pp) then begin
      if (pp is TtsPostProcessorFillPattern) then with (pp as TtsPostProcessorFillPattern) do begin
        if OwnsPattern then
          DelReference(ltsObjTypeImage, Pattern);
      end;
      DelReference(ltsObjTypePostProcessor, pp);
      FreeAndNil(pp);
    end else
      result := LastErrorCode;
  except
    on ex: Exception do begin
      SetLastError(ex);
      result := LastErrorCode;
    end;
  end;
  {$IFDEF DEBUG}utlLogger.Log(nil, 'ltsPostProcessorDestroy=%d', [Integer(result)]);{$ENDIF}
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//TltsPostProcessorCustom///////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TltsPostProcessorCustom.Execute(const aChar: TtsChar; const aImage: TtsImage): Boolean;
begin
  result := inherited Execute(aChar, aImage);
  if result then begin
    AddReference(ltsObjTypeChar, aChar);
    AddReference(ltsObjTypeImage, aImage);
    try
      fData.execute(aChar, aImage, fData.args);
    finally
      DelReference(ltsObjTypeChar, aChar);
      DelReference(ltsObjTypeImage, aImage);
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TltsPostProcessorCustom.Create(const aContext: TtsContext; const aData: TltsPostProcessorCustomData);
begin
  inherited Create(aContext);
  fData := aData;
end;

end.

